/*
Stolyarevskiy Sergey, http://acobw.narod.ru

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation.
	
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
		
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

23.03.2005 v1.0
27.03.2005 v1.1 Corrected: -incorrect work after using a scrap of edges
10.04.2005 v1.2 optimization
24.04.2005 v1.3 some changes to algorithm
06.11.2005 v1.4 corrected mistakes
14.03.2007 v1.5 optimization


*/
#include <windows.h>
#include <stdio.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "filter.h"

int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
int StartProc(FilterActivation *fa, const FilterFunctions *ff);
int InitProc(FilterActivation *fa, const FilterFunctions *ff);
int EndProc(FilterActivation *fa, const FilterFunctions *ff);

void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

ScriptFunctionDef _func_defs[]={
    { (ScriptFunctionPtr)ScriptConfig, "Config", "0i" },
    { NULL },
};

CScriptObject ndf_obj={
    NULL, _func_defs
};

typedef struct MyFilterData {
    unsigned char   *old_data;
	_int64 NDelit;
	int NLevel;
	unsigned char SDelit;
    long size;
} MyFilterData;

struct FilterDefinition filterDef_ = {

    NULL, NULL, NULL,									// next, prev, module
    "NDF, version 1.5",									// name
    "Nonlinear dynamic noise a supressor, version 1.5", // desc
    "Stolyarevskiy Sergey",								// maker
    NULL,												// private_data
    sizeof(MyFilterData),								// inst_data_size
    InitProc,											// InitProc
    NULL,												// deinitProc
    RunProc,											// runProc
	NULL,												// ParamProc
    ConfigProc,											// configProc
    StringProc,											// stringProc,
    StartProc,											// startProc
    EndProc,											// endProc
    &ndf_obj,											// script_obj
    FssProc,											// fssProc
};


extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_ = ff->addFilter(fm, &filterDef_, sizeof(FilterDefinition))))
        return 1;

    vdfd_ver    = VIRTUALDUB_FILTERDEF_VERSION;
    vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_);
}

//-----------------------------------------------------------------------------------------------------------------
int RunProc(const FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
  Pixel32 *src, *dst, *old;
 _int64 Zero=0x0000000000000000;
 _int64 ND=mfd->NDelit;
 _int64 SD=mfd->SDelit;
	src=(Pixel32 *)fa->src.data;
    dst=(Pixel32 *)fa->dst.data;
	old=(Pixel32 *)mfd->old_data;
_asm{
	movq mm0, Zero			//
	movq mm5, ND			// 64
	movd mm7, SD
	}
int h = fa->src.h;
do {
int w = fa->src.w;						//0,1,2,3, ,5, ,7
do {
int Pt=*src++;
int P_old=*old++;
_asm{			movd mm1, Pt			// 
				punpcklbw mm1, mm0		//   
				movd mm2, P_old			//  
				punpcklbw mm2, mm0		//   
				psubsw mm2, mm1			//n-T
				movq mm3, mm2			//   n-T
				psrlw mm2, 1			//  2
				pmullw mm2, mm2			//   ((Tn-T)/2)^2
				movq mm4, mm2
				psrlq mm4, 16			//  G
				paddusw mm2, mm4		// G+B
				psrlq mm4, 16			//  R
				paddusw mm2, mm4		// R+(G+B)
				psraw mm2, 1			//  2
				movq mm4, mm5
				psubusw mm4, mm2		//64-((Tn-T)/2)^2
				punpcklwd mm4, mm4
				punpckldq mm4, mm4
				pmullw mm3, mm4			//(n-T)*(64-((Tn-T)/2)^2)
				psraw mm3, mm7			//cor=cor/64/2
		//		movq mm1, ND			//  
		//		psllq mm3, 4
				paddsw mm1, mm3			//+r/2
		//		psubsw mm1, mm3			//-r/2
				packuswb mm1, mm0		//
	}

int T;
_asm{movd T, mm1}
*dst++=T;	
     } while(--w);
        src = (Pixel32 *)((char *)src + fa->src.modulo);
        dst = (Pixel32 *)((char *)dst + fa->dst.modulo);
		old = (Pixel32 *)((char *)old + fa->src.modulo);
} while(--h);
_asm{emms}
memcpy (mfd->old_data,fa->dst.data,mfd->size);
	return 0;
}

void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
    const char *modes[5]={
	    "  -  Minimum",
        "  -  Weakly",
        "  -  Normal",
		"  -  Powerfully",
		"  -  Maximum"
	};
   MyFilterData *mfd = (MyFilterData *)fa->filter_data;
		strcpy(str, modes[mfd->NLevel]);
}

int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	if (++mfd->NLevel>4) mfd->NLevel=0;
	Beep((mfd->NLevel+2)*600,20);
	return 0;
}

int StartProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
mfd->size = fa->src.Size ();
if (!(mfd->old_data = new unsigned char [fa->src.Size()]))
        return 1;
 memset (mfd->old_data, 0, mfd->size);

switch (mfd->NLevel){
case 0:
	mfd->NDelit=0x0000000000000010;
	mfd->SDelit=5;
	break;
case 1:
	mfd->NDelit=0x0000000000000020;
	mfd->SDelit=6;
	break;
case 2:
	mfd->NDelit=0x0000000000000040;
	mfd->SDelit=7;
	break;
case 3:
	mfd->NDelit=0x0000000000000080;
	mfd->SDelit=8;
	break;
case 4:
	mfd->NDelit=0x0000000000000100;
	mfd->SDelit=9;
	break;
}
    return 0;
}

int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    mfd->NLevel = 1;
    return 0;
}


void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    mfd->NLevel=argv[0].asInt();
}

bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    _snprintf(buf, buflen,"Config(%d)",mfd->NLevel);
    return true;
}
int EndProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    delete[] mfd->old_data;
    mfd->old_data = NULL;
    return 0;
}